-- Time-stamp: <Fri Oct 21 2005 13:55:58 Stardate: [-29]4737.48 hwloidl>
--
-- Interpreter for the Dummy Language from Appel's compiler book (Chapter 1)
---------------------------------------------------------------------------

module DLInterp where

import DLAbsSyn
import DLParse
import Data.FiniteMap

import Debug.Trace(trace)

type Env = FiniteMap Identifier Val
type Trace = String
type Val = Int
{-
data Val = IVal Int
         | BVal Bool
         | SVal String
-}


emptyEnv :: Env
emptyEnv = emptyFM

emptyTrace :: Trace
emptyTrace = ""

class Evaluator x where
 eval :: Env -> Trace -> x -> (Val, Trace, Env)
 prints :: String -> Env -> [x] -> (Trace, Env)
 prints s t [] = (s,t)
 prints s t (e1:es) = let 
                        (v,q',t') = eval t s e1
                      in
                        prints (q'++(show v)++"\n") t' es 

instance Evaluator Stm where
 -- eval :: Env -> Stm -> (Val, Env)
 eval t q (CompoundStm s1 s2) = let 
                                (v',q',t') = eval t q s1
                               in
                                eval t' q' s2
 eval t q (AssignStm id e) = let 
			      (v,q',t') = eval t q e
                              t'' = addToFM t' id v 
                          in
                            (v,q',t'')
 eval t q (PrintStm xs) = let 
                          (q',t') = prints "" t xs
                        in
                          (0,q++q',t')
 eval t q (IfStm e s1 s2) = let 
                              (v,q',t') = eval t q e
                            in
                              if (v==0)
                                then eval t' q' s2
                                else eval t' q' s1
 eval t q (WhileStm e s) =  let 
                              (v,q',t') = eval t q e
                            in
                              if (v==0)
                                then (0, q', t')
                                else let 
                                       (v'',q'',t'') = eval t' q' s
                                     in
                                       eval t'' q''  (WhileStm e s)
 
instance Evaluator Exp where
 --eval :: Env -> Expr -> (Val, Env)
 eval t q (IdExp id) = case (lookupFM t id) of
                        Nothing -> error ("Unknown identifier " ++ id)
                        (Just v) -> (v,q,t)

 eval t q (NumExp n) = (n, q, t)

 eval t q (OpExp e1 op e2) = let
                             (v1,q1,t1) = eval t q e1
                             (v2,q2,t2) = eval t1 q1 e2
                           in
                             case op of
                                OpPlus -> (v1+v2, q2, t2)
                                OpMinus -> (v1-v2, q2, t2)
                                OpTimes -> (v1*v2, q2, t2)
                                OpEq -> (if (v1==v2) then 1 else 0, q2, t2)
                                OpLt -> (if (v1<v2) then 1 else 0, q2, t2)
                                OpAnd -> (v1*v2, q2, t2)
                                _ -> error ((show op) ++" not yet implemented")
 eval t q (EsepExp s e) = let
                          (v',q',t') = eval t q s
                        in 
                          eval t' q' e

